#ifdef _WINDOWS
#define WINDOWS
#endif

#include "ac_plugin.h"
#include <stdio.h>
#include <string.h>

static void ddhBresLine( bool **pixels, int sizex, int sizey, int x1, int y1, int x2, int y2 );

//---------------------------------------------------------------------------------------
// Export UVs
static void ddhexportuv_go( int argc, const char* argv[] )
{

   // Get selection
   List *selsurfs = ac_selection_get_whole_surfaces_all();
   if( selsurfs == NULL )
   {
      message_dialog( "exportuv: no surfaces selected" );
      return;
   }

   // Get params
   if( argc < 4 )
   {
      list_free( &selsurfs );
      message_dialog( "exportuv: insufficient parameters passed to exportuv plugin" );
      return;
   }

   int sizex, sizey;
   bool white_on_black;
   if( sscanf( argv[1], "%d", &( sizex )) != 1 )
   {
      list_free( &selsurfs );
      message_dialog( "exportuv: invalid width specified (%s)", argv[1] );
      return;
   }
   if( sscanf( argv[2], "%d", &( sizey )) != 1 )
   {
      list_free( &selsurfs );
      message_dialog( "exportuv: invalid height specified (%s)", argv[2] );
      return;
   }
   white_on_black = ( strcmp( "0", argv[3] ) != 0 );

   // Limit sizes
   if( sizex < 1 || sizex > 16384 )
   {
      list_free( &selsurfs );
      message_dialog( "exportuv: width must be between 1 and 16384 pixels (%d)", sizex );
      return;
   }
   if( sizey < 1 || sizey > 16384 )
   {
      list_free( &selsurfs );
      message_dialog( "exportuv: height must be between 1 and 16384 pixels (%d)", sizey );
      return;
   }

   // Open the file
   char *filter[] = { "BMP files", ".bmp", "All files", "*", NULL };
   char *fname = ac_get_save_filename( "Export UV map to a BMP file", filter );
   if(( fname == NULL ) || ( *fname =='\0' ))
   {
      list_free( &selsurfs );
      return;
   }

   FILE *fp = fopen( fname, "wb" );
   if( fp == NULL )
   {
      message_dialog( "exportuv: can't open file '%s' for writing", fname );
      list_free( &selsurfs );
      return;
   }

   // Create the pixel array
   bool **pixels = new bool*[sizex];
   for( int ix = 0; ix < sizex; ++ix )
   {
      pixels[ix] = new bool[sizey];
      for( int iy = 0; iy < sizey; ++iy )
      {
         pixels[ix][iy] = false;
      }
   }

   // Draw the lines
   float fsizex = ( float )sizex - 1.0f;
   float fsizey = ( float )sizey - 1.0f;
   for( List *ps = selsurfs; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      int ctr = 0;
      for( List *psv = s->vertlist; psv != NULL; psv = psv->next )
      {
         SVertex *sv = ( SVertex * )( psv->data );
         SVertex *sv2;
         ++ctr;
         if( ctr == s->numvert )
         {
            // Draw a line back to zero
            sv2 = ( SVertex * )( s->vertlist->data );
         }
         else
         {
            // Draw a line to the next coords
            sv2 = ( SVertex * )( psv->next->data );
         }
         ddhBresLine( pixels, sizex, sizey,
                      ( int )( sv->tx * fsizex ), ( int )( sv->ty * fsizey ),
                      ( int )( sv2->tx * fsizex ), ( int )( sv2->ty * fsizey ));
      }
   }

   // Calculate the pixel size
   int fsize = sizex * sizey;
   int i;
   short s;

   // Write the BMP header
   fwrite( "BM", 2, 1, fp );
   // File size
   i = fsize * 3 + 40 + 14;
   fwrite( &i, 4, 1, fp );
   // Reserved
   i = 0;
   fwrite( &i, 4, 1, fp );
   // Offset to data
   i = 54;
   fwrite( &i, 4, 1, fp );

   // Write the Image header
   // Header size
   i = 40;
   fwrite( &i, 4, 1, fp );
   // Image width
   fwrite( &sizex, 4, 1, fp );
   // Image height
   fwrite( &sizey, 4, 1, fp );
   // Planes
   s = 1;
   fwrite( &s, 2, 1, fp );
   // BPP
   s = 24;
   fwrite( &s, 2, 1, fp );
   // Compression
   i = 0;
   fwrite( &i, 4, 1, fp );
   // Image size (may be zero for uncompressed images)
   i = fsize * 3;
   fwrite( &i, 4, 1, fp );
   // X ppm
   i = 0;
   fwrite( &i, 4, 1, fp );
   // Y ppm
   fwrite( &i, 4, 1, fp );
   // Color map entries
   fwrite( &i, 4, 1, fp );
   // Significant colors
   fwrite( &i, 4, 1, fp );

   unsigned char bg[3], fg[3];
   if( white_on_black )
   {
      bg[0] = 0x00;
      bg[1] = 0x00;
      bg[2] = 0x00;
      fg[0] = 0xff;
      fg[1] = 0xff;
      fg[2] = 0xff;
   }
   else
   {
      bg[0] = 0xff;
      bg[1] = 0xff;
      bg[2] = 0xff;
      fg[0] = 0x00;
      fg[1] = 0x00;
      fg[2] = 0x00;
   }
   // Determine the number of zero-bytes to pad each row to get a multiple of 4
   int rowpad = sizex % 4;
   int zero = 0;
   // For better performance, write to memory first, dump 1 row at a time
   unsigned char *pixelbuffer = new unsigned char[sizex * 3];
   set_progress_bar( 0 );
   for( int iy = 0; iy < sizey; ++iy )
   {
      unsigned char *p = pixelbuffer;
      for( int ix = 0; ix < sizex; ++ix )
      {
         if( pixels[ix][iy] )
         {
            p[0] = fg[0];
            p[1] = fg[1];
            p[2] = fg[2];
         }
         else
         {
            p[0] = bg[0];
            p[1] = bg[1];
            p[2] = bg[2];
         }
         p += 3;
      }
      set_progress_bar(( int )(( float )iy / fsizey * 100.0f ));
      fwrite( pixelbuffer, sizex * 3, 1, fp );
      if( rowpad > 0 )
      {
         fwrite( &zero, rowpad, 1, fp );
      }
   }
   set_progress_bar( 0 );
   delete[] pixelbuffer;
   /*
   for( int iy = 0; iy < sizey; ++iy )
   {
      for( int ix = 0; ix < sizex; ++ix )
      {
         if( pixels[ix][iy] )
         {
            fwrite( &fg, 3, 1, fp );
         }
         else
         {
            fwrite( &bg, 3, 1, fp );
         }
      }
   }
   */

   // Cleanup
   for( int ix = 0; ix < sizex; ++ix )
   {
      delete[] pixels[ix];
   }
   delete[] pixels;
   fclose( fp );
   list_free( &selsurfs );

}
//---------------------------------------------------------------------------------------
static void ddhBresLine( bool **pixels, int sizex, int sizey, int x1, int y1, int x2, int y2 )
{

   // Line drawing source code adapted from algorithm posted by Tom Ootjers on gamedev.net, 2001

   int deltax = abs(x2 - x1);        // The difference between the x's
   int deltay = abs(y2 - y1);        // The difference between the y's
   int x = x1;                       // Start x off at the first pixel
   int y = y1;                       // Start y off at the first pixel
   int xinc1, xinc2, yinc1, yinc2, den, num, numadd, numpixels;

   if (x2 >= x1)                 // The x-values are increasing
   {
      xinc1 = 1;
      xinc2 = 1;
   }
   else                          // The x-values are decreasing
   {
      xinc1 = -1;
      xinc2 = -1;
   }

   if (y2 >= y1)                 // The y-values are increasing
   {
      yinc1 = 1;
      yinc2 = 1;
   }
   else                          // The y-values are decreasing
   {
      yinc1 = -1;
      yinc2 = -1;
   }

   if (deltax >= deltay)         // There is at least one x-value for every y-value
   {
      xinc1 = 0;                  // Don't change the x when numerator >= denominator
      yinc2 = 0;                  // Don't change the y for every iteration
      den = deltax;
      num = deltax / 2;
      numadd = deltay;
      numpixels = deltax;         // There are more x-values than y-values
   }
   else                          // There is at least one y-value for every x-value
   {
      xinc2 = 0;                  // Don't change the x for every iteration
      yinc1 = 0;                  // Don't change the y when numerator >= denominator
      den = deltay;
      num = deltay / 2;
      numadd = deltax;
      numpixels = deltay;         // There are more y-values than x-values
   }

   for (int curpixel = 0; curpixel <= numpixels; curpixel++)
   {
      if( x > -1 && y > -1 && x < sizex && y < sizey )
      {
         pixels[x][y] = true;   // Draw the current pixel
      }
      num += numadd;              // Increase the numerator by the top of the fraction
      if (num >= den)             // Check if numerator >= denominator
      {
         num -= den;               // Calculate the new numerator value
         x += xinc1;               // Change the x as appropriate
         y += yinc1;               // Change the y as appropriate
      }
      x += xinc2;                 // Change the x as appropriate
      y += yinc2;                 // Change the y as appropriate
   }

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/exportuv.tcl\"", path );
      ac_add_tools_menu_item( "Export UV map...", "ddhexportuv_dialog",
                              "Export UV map for selected surfaces" );
      ac_add_command_full( "ddhexportuv_go", ( void * )ddhexportuv_go,
                            0, "argv", "<arg1> <arg2>",
                           "Export UV map" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Export UV map - version 1.0.1 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Export UV map - Dennis Hawthorne - Export selected surfaces' UV coordinates to a BMP file." );
}
//---------------------------------------------------------------------------------------
